//ACTIVE_ACCOUNT must be an object of type Account
//REPORT_COUNT must be of type theCounter( mailCount )
//LAST_MAIL must be of type theLastMail( sender, subject, date )
//FAIL( ACTIVE_ACCOUNT ) is called on errors
//DONE( ACTIVE_ACCOUNT ) is called on finished communication
//public commands are void checkMail() and void kill()
//please call kill() on removing

function mkiiImapClient( ACTIVE_ACCOUNT, REPORT_COUNT, LAST_MAIL, FAIL, DONE ) {
	var serverConnection;
	var state = "";
	var stateCount = 0;
	var ready = true;
	var timeoutCounter;
	
	function debug( STRING ) {
		_debug( "mkiiImapClient : " + STRING );
	}
	
	this.checkMail = function() {
		debug( "checkMail()" );
		
		this.kill();
		connect();
	};
	
	this.kill = function() {
		debug( "kill()" );
	
		if ( !ready ) {
			cancel();
		}
	};
	
	function retry() {
		debug( "retry()" );
		
		if ( ACTIVE_ACCOUNT.newLine == "\n" ) {
			debug( "communication error. retrying with different line ending" );
			cancel();
			ACTIVE_ACCOUNT.newLine = "\r\n";
			setTimeout( connect, 5000 );
		} else {
			ACTIVE_ACCOUNT.newLine = "\n";
			cancel();
			DONE( ACTIVE_ACCOUNT );
			FAIL( ACTIVE_ACCOUNT );
		}
	}
	
	function failure() {
		ACTIVE_ACCOUNT.newLine = "\n";
		FAIL( ACTIVE_ACCOUNT );
		logout();
	}
	
	function forceDisconnect() {
		ACTIVE_ACCOUNT.newLine = "\n";
		cancel();
		DONE( ACTIVE_ACCOUNT );
		FAIL( ACTIVE_ACCOUNT );
	}
	
	function connect() {
		debug( "connect()" );
	
		//timeoutCounter = setTimeout( retry, 30000 );
		stateCount = 0;
		ready = false;
		
		if ( ACTIVE_ACCOUNT.SSLEnabled ) {
			debug( "/usr/bin/openssl s_client -connect " + ACTIVE_ACCOUNT.Hostname + ":" + ACTIVE_ACCOUNT.PortNumber );
			serverConnection = widget.system( "/usr/bin/openssl s_client -connect " + ACTIVE_ACCOUNT.Hostname + ":" + ACTIVE_ACCOUNT.PortNumber, imapEndHandler );
		} else {
			debug( "/usr/bin/telnet " + ACTIVE_ACCOUNT.Hostname + " " + ACTIVE_ACCOUNT.PortNumber );
			serverConnection = widget.system( "/usr/bin/telnet " + ACTIVE_ACCOUNT.Hostname + " " + ACTIVE_ACCOUNT.PortNumber, imapEndHandler );
		}
		
		serverConnection.onreaderror = errorHandler;
		serverConnection.onreadoutput = imapHandler;
		timeoutCounter = setTimeout( retry, 5000 );
	}
	
	function login() {
		debug( "login()" );
		
		stateCount = 1;
		//debug( "sending : a01 LOGIN " + ACTIVE_ACCOUNT.Username + " " + ACTIVE_ACCOUNT.Password );
		debug( "sending : a01 LOGIN " + ACTIVE_ACCOUNT.Username + " PASSWORD" );
		serverConnection.write( "a01 LOGIN " + ACTIVE_ACCOUNT.Username + " " + ACTIVE_ACCOUNT.Password + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( retry, 5000 );
	}
	
	function cramMD5() {
		debug( "cramMD5()" );
	
		stateCount = 2;
		debug( "sending : a01 AUTHENTICATE CRAM-MD5" );
		serverConnection.write( "a01 AUTHENTICATE CRAM-MD5" + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( retry, 5000 );
	}
	
	function cramMD5Response( SERVER_RESPONSE ) {
		debug( "cramMD5Response()" );
		
		stateCount = 3;
		var clientResponse = base64Encode( ACTIVE_ACCOUNT.Username + " " + hex_hmac_md5( ACTIVE_ACCOUNT.Password, SERVER_RESPONSE ) );
		debug( "sending : " + clientResponse );
		serverConnection.write( clientResponse + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( retry, 5000 );
	}
		
	function select() {
		debug( "select()" );
	
		stateCount = 4;
		debug( "sending : a02 SELECT \"INBOX\"" );
		serverConnection.write( "a02 SELECT \"INBOX\"" + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( retry, 5000 );
	}
	
	function fetch() {
		debug( "fetch()" );
		
		stateCount = 5;
		state = "";
		debug( "sending : a03 FETCH 1:* (UID FLAGS)" );
		serverConnection.write( "a03 FETCH 1:* (UID FLAGS)" + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( failure, 10000 );
	}
	
	function fetchHeader( MESSAGE ) {
		debug( "fetchHeader( " + MESSAGE + " )" );
		
		stateCount = 6;
		state = "";
		debug( "sending : a04 FETCH " + MESSAGE + " RFC822.HEADER" );
		serverConnection.write( "a04 FETCH " + MESSAGE + " RFC822.HEADER" + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( failure, 10000 );
	}
	
	function logout() {
		debug( "logout()" );
		
		stateCount = 7;
		debug( "sending : a05 LOGOUT" );
		serverConnection.write( "a05 LOGOUT" + ACTIVE_ACCOUNT.newLine );
		timeoutCounter = setTimeout( failure, 10000 );
	}
	
	function cancel() {
		debug( "cancel()" );
		
		serverConnection.cancel();
		ready = true;
		clearTimeout( timeoutCounter );
	}
	
	function imapEndHandler( RESULT ) {
		debug( "imapEndHandler( " + RESULT + " )" );
		
		ready = true;
		clearTimeout( timeoutCounter );
		
		if ( stateCount < 7 ) {
			FAIL( ACTIVE_ACCOUNT );
		} else {
			DONE( ACTIVE_ACCOUNT );
		}
	}
	
	function errorHandler( RESULT ) {
		debug( "errorHandler( " + RESULT + " )" );
	}
		
	function imapHandler( RESULT ) {
		debug( "imapHandler( RESULT )" );
		debug( "imapHandler : " + stateCount );
		
		clearTimeout( timeoutCounter );
/*	
		if ( stateCount < 4 ) {
			timeoutCounter = setTimeout( retry, 5000 );
		} else {
			timeoutCounter = setTimeout( failure, 10000 );
		}
*/				
		switch ( stateCount ) {
			case 0:	connectHandler( RESULT );
					break;
			case 1:	loginHandler( RESULT );
					break;
			case 2: cramMD5Handler( RESULT );
					break;
			case 3: cramMD5ResponseHandler( RESULT );
					break;
			case 4:	selectHandler( RESULT );
					break;
			case 5:	fetchHandler( RESULT );
					break;
			case 6:	fetchHeaderHandler( RESULT );
					break;
			case 7:	logoutHandler( RESULT );
					break;
			default:
		}
	}

	function connectHandler( RESULT ) {
		debug( "connectHandler( " + RESULT + " )" );
		
		if ( RESULT.indexOf( "* OK" ) != -1 ) {
			debug( "connectHandler : we are connected" );
			
			if ( ACTIVE_ACCOUNT.AuthenticationScheme == "CRAM-MD5" ) {
				cramMD5();
			} else {
				login();
			}
		}
	}

	function loginHandler( RESULT ) {
		debug( "loginHandler( " + RESULT + " )" );
		
		if ( ( RESULT.indexOf( "a01 BAD" ) != -1 ) || ( RESULT.indexOf( "a01 NO" ) != -1 ) ) {
			debug( "fixMe : we are unable to login" );
			failure();
		}
		
		if ( RESULT.indexOf( "a01 OK" ) != -1 ) {
			debug( "loginHandler : we are connected" );
			select();
		}
	}

	function cramMD5Handler( RESULT ) {
		debug( "cramMD5Handler( " + RESULT + " )" );
		
		if ( RESULT.indexOf( "+ " ) != -1 ) {
			cramMD5Response( base64Decode( RESULT.split( " " )[1] ) );
		} else {
			debug( "fixMe : no CRAM-MD5 ?" );
			failure();
		}
	}

	function cramMD5ResponseHandler( RESULT ) {
		debug( "cramMD5ResponseHandler : " + RESULT );
		
		if ( ( RESULT.indexOf( "a01 BAD" ) != -1 ) || ( RESULT.indexOf( "a01 NO" ) != -1 ) ) {
			debug( "fixMe : we are unable to cram-md5" );
			failure();
		}
		
		if ( RESULT.indexOf( "a01 OK" ) != -1 ) {
			debug( "loginHandler : we are connected" );
			select();
		}
	}

	function selectHandler( RESULT ) {
		debug( "selectHandler( " + RESULT + " )" );
		
		if ( ( RESULT.indexOf( "a02 BAD" ) != -1 ) || ( RESULT.indexOf( "a02 NO" ) != -1 ) ) {
			debug( "fixMe : we are unable to select" );
			failure();
		}
		
		if ( RESULT.indexOf( "a02 OK" ) != -1 ) {
			debug( "selectHandler : successfull selected" );
			fetch();
		}
	}

	function fetchHandler( RESULT ) {
		debug( "fetchHandler( " + RESULT + " )" );
		state += RESULT;
		
		if ( ( RESULT.indexOf( "a03 BAD" ) != -1 ) || ( RESULT.indexOf( "a03 NO" ) != -1 ) ) {
			debug( "fixMe : we are unable to fetch" );
			failure();
		}
		
		if ( state.indexOf( "a03 OK" ) != -1 ) {
			debug( "fetchHandler : successfull fetched" );
			state = state.replace(/\r/g, '' );
			var splittedState = state.split( '\n' );
			var newMails = 0;
			
			for ( var i = 0; i < splittedState.length; i++ ) {
				if ( ( splittedState[i].indexOf( "UID" ) != -1 ) && ( splittedState[i].indexOf( "\\Seen" ) == -1 ) ) {
					newMails++;
				}
			}
			
			debug( "fetchHandler : reporting " + newMails + " new mails" );
			REPORT_COUNT( newMails );
			
			if ( newMails > 0 ) {
				/* the most recent message is the one with no \seen and highest UID */
				state = state.replace(/\r/g, '' );
				var temp = state.split( '\n' );
				var latestMailID = 0;
				var latestMail = 0;
				
				for ( var i = 0; i < temp.length; i++ ) {
					if ( temp[i].indexOf( "\\Seen" ) == -1 ) {
						if ( parseInt( temp[i].split( "UID " )[1] ) > latestMailID ) {
							latestMailID = parseInt( temp[i].split( "UID " )[1] );
							latestMail = parseInt( temp[i].split( "* " )[1] );
							debug( latestMailID + " - " + latestMail );
						}
					}
				}
			
				debug( "fetchHandler : try to fetch " + latestMail );
				fetchHeader( latestMail );
			} else {
				//LAST_MAIL( null, null, null );
				logout();
			}
		}
	}
	
	function fetchHeaderHandler( RESULT ) {
		debug( "fetchHeaderHandler( " + RESULT + " )" );
		state += RESULT;
		
		if ( ( RESULT.indexOf( "a04 BAD" ) != -1 ) || ( RESULT.indexOf( "a04 NO" ) != -1 ) ) {
			debug( "fixMe : we are unable to fetch the Header" );
			failure();
		}
		
		if ( state.indexOf( "a04 OK" ) != -1 ) {
			debug( "fetchHeaderHandler : successfull fetched" );
			state = state.replace(/\r/g, '' );
			var temp = state.split( '\n' );			
			var from = "";
			var subject = "";
			var date = "";
			
			for ( var i = 0; i < temp.length; i++ ) {
				if ( temp[i].indexOf( "From:" ) != -1 ) {
					from = temp[i].split( "From:" )[1];
					var n = 1;
					
					while ( ( temp[i+n].charAt( 0 ) == " " ) || temp[i+n].charAt( 0 ) == "	" ) {
						from += temp[i+n];
						n++;
					}
					
					from = from.replace( /"/g, '\\\"' );
				}
				
				if ( temp[i].indexOf( "Subject:" ) != -1 ) {
					subject = temp[i].split( "Subject:" )[1];
					var n = 1;
					
					while ( ( temp[i+n].charAt( 0 ) == " " ) || temp[i+n].charAt( 0 ) == "	" ) {
						subject += temp[i+n];
						n++;
					}
					
					subject = subject.replace( /"/g, '\\\"' );
				}
				
				if ( temp[i].indexOf( "Date: " ) != -1 ) {
					date = temp[i].split( "Date: " )[1];
				}
			}
			
			debug( "eMail from : " + from + "\nSubject : " + subject + "\nReceived on : " + date );
			LAST_MAIL( from, subject, date );
			logout();
		}
	}

	function logoutHandler( RESULT ) {
		debug( "logoutHandler( " + RESULT + " )" );
		
		if ( ( RESULT.indexOf( "a05 BAD" ) != -1 ) || ( RESULT.indexOf( "a05 NO" ) != -1 ) ) {
			debug( "fixMe : unable to logout" );
			forceDisconnect();
		}
		
		if ( RESULT.indexOf( "* BYE" ) != -1 ) {
			debug( "logoutHandler : we are out" );
		}
	}
}
